// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_ALBERTAGRID_HH
#define DUNE_ALBERTAGRID_HH
#include "albertagrid/agrid.hh"
#include "albertagrid/gridfactory.hh"
#include "albertagrid/persistentcontainer.hh"
#endif
