// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

#ifndef DUNE_GRID_IO_FILE_VTK_HH
#define DUNE_GRID_IO_FILE_VTK_HH

/** \file
    \brief Convenience header which includes all available VTK writers
 */

#include "vtk/boundarywriter.hh"
#include "vtk/subsamplingvtkwriter.hh"
#include "vtk/vtksequencewriter.hh"
#include "vtk/vtkwriter.hh"
#include "vtk/volumewriter.hh"

#endif // DUNE_GRID_IO_FILE_VTK_HH
