// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

#include <config.h>

#include <iostream>

#include <dune/geometry/virtualrefinement.hh>

#include <dune/grid/io/file/vtk/vtkwriter.hh>

/*
   #include "gridcheck.cc"
   #include "checkgeometryinfather.cc"
   #include "checkintersectionit.cc"
   #include "checkpartition.cc"
 */

template <class Grid>
void test(Grid& grid)
{}
