// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
/** \file
 * \brief Contains <tt>#undef</tt>s for all preprocessor macros
 * defined by UG.
 *
 * This file is created automatically by the perl script <tt>undefAllMacros.pl</tt>.
 */

#undef ABS
#undef ABSDIFF
#undef ADDPATTERN
#undef ADDPATTERN_LEN
#undef ADDPATTERN_SHIFT
#undef __ALGEBRA__
#undef __ANISOTROPIC__
#undef APOS_TO_AID
#undef AREA_OF_ELEMENT
#undef AREA_OF_ELEMENT_2D
#undef AREA_OF_ELEMENT_3D
#undef AREA_OF_ELEMENT_BND
#undef AREA_OF_HEXAHEDRON
#undef AREA_OF_PRISM
#undef AREA_OF_PYRAMID
#undef AREA_OF_QUADRILATERAL
#undef AREA_OF_REF
#undef AREA_OF_REF_2D
#undef AREA_OF_REF_3D
#undef AREA_OF_REF_BND
#undef AREA_OF_TETRAHEDRON
#undef AREA_OF_TRIANGLE
#undef ASSERT
#undef ATTR
#undef ATTR_TO_GLEVEL
#undef AUXEDGE
#undef AUXEDGE_LEN
#undef AUXEDGE_SHIFT
#undef BHM_OK
#undef BITWISE_TYPE
#undef BLOCK_NOT_DEFINED
#undef __BLOCK_VECTOR_DESC__
#undef BLOCKVECTOR_OFFSET
#undef BND_SIZE_TAG
#undef BOOL_2_NF
#undef BOOL_2_TF
#undef BOOL_2_YN
#undef BOTTOMLEVEL
#undef Broadcast
#undef BV_CYC
#undef BVD_DEC_LAST_ENTRY
#undef BVD_DISCARD_LAST_ENTRY
#undef BVD_GET_ENTRY
#undef BVD_INC_LAST_ENTRY
#undef BVD_INIT
#undef BVD_INIT_SEQ_READ
#undef BVD_IS_SUB_BLOCK
#undef BVD_MAX_ENTRIES
#undef BVD_NR_ENTRIES
#undef BVDOWNBV
#undef BVDOWNBVEND
#undef BVDOWNBVLAST
#undef BVDOWNTYPE
#undef BVDOWNTYPE_LEN
#undef BVDOWNTYPE_SHIFT
#undef BVDOWNVECTOR
#undef BVD_PUSH_ENTRY
#undef BVD_READ_NEXT_ENTRY
#undef BVD_SET_ENTRY
#undef BVENDVECTOR
#undef BVFIRSTVECTOR
#undef BV_GEN
#undef BV_GEN_C
#undef BV_GEN_F
#undef BV_GEN_L
#undef BV_IS_DIAG_BV
#undef BV_IS_EMPTY
#undef BV_IS_GC
#undef BV_IS_LEAF_BV
#undef BVLASTVECTOR
#undef BVLEVEL
#undef BVLEVEL_LEN
#undef BVLEVEL_SHIFT
#undef BVNUMBER
#undef BVNUMBEROFVECTORS
#undef BVORIENTATION
#undef BVORIENTATION_LEN
#undef BVORIENTATION_SHIFT
#undef BVPD_CONFIG
#undef BVPD_CONVEX
#undef BVPD_MIDPOINT
#undef BVPD_NAME
#undef BVPD_NCOEFFF
#undef BVPD_NPARTS
#undef BVPD_NSUBDOM
#undef BVPD_NUSERF
#undef BVPD_RADIUS
#undef BVPD_S2P
#undef BVPD_S2P_PTR
#undef BVPRED
#undef BVSUCC
#undef BVTVTYPE
#undef BVTVTYPE_LEN
#undef BVTVTYPE_SHIFT
#undef BVUSERDATA
#undef CANNOT_INIT_PROBLEM
#undef CAST_EDVECTOR
#undef CAST_EVECTOR
#undef CAST_NVECTOR
#undef CAST_SVECTOR
#undef CAT
#undef CAT3
#undef CE_FREE
#undef CE_INIT
#undef CE_INIT_UNUSED
#undef CE_LOCKED
#undef CENTERNODE
#undef CENTER_NODE_INDEX
#undef CENTER_NODE_INDEX_TAG
#undef CENTERTYPE
#undef CE_USED
#undef CEXTRA
#undef CEXTRA_LEN
#undef CEXTRA_SHIFT
#undef C_FRONTEND
#undef CHECK
#undef CLASS_OF_RULE
#undef CLEAR_FLAG
#undef CLEAR_VEC_COLLECT__FLAG
#undef CLEAR_VEC_CONS__FLAG
#undef CMATRIX0
#undef CMATRIX1
#undef COARSEN
#undef COARSEN_LEN
#undef COARSEN_SHIFT
#undef __COMMANDS__
#undef CONCAT
#undef CONCAT3
#undef CONCAT3_AUX
#undef CONCAT4
#undef CONCAT4_AUX
#undef CONCAT5
#undef CONCAT5_AUX
#undef CONCAT_AUX
#undef CONTEXT
#undef ControlWord
#undef COOBJ
#undef COORDINATE_TO_KEY
#undef COPY_COORDINATES_HEXAHEDRON
#undef COPY_COORDINATES_PRISM
#undef COPY_COORDINATES_PYRAMID
#undef COPY_COORDINATES_QUADRILATERAL
#undef COPY_COORDINATES_TETRAHEDRON
#undef COPY_COORDINATES_TRIANGLE
#undef COPY_CORNER_COORDINATES
#undef COPY_SC_TO_SH
#undef CORNER
#undef CORNER_COORDINATES
#undef CORNER_COORDINATES_HEXAHEDRON
#undef CORNER_COORDINATES_PRISM
#undef CORNER_COORDINATES_PYRAMID
#undef CORNER_COORDINATES_QUADRILATERAL
#undef CORNER_COORDINATES_TETRAHEDRON
#undef CORNER_COORDINATES_TRIANGLE
#undef CORNER_OF_EDGE
#undef CORNER_OF_EDGE_PTR
#undef CORNER_OF_EDGE_REF
#undef CORNER_OF_EDGE_TAG
#undef CORNER_OF_OPPEDGE
#undef CORNER_OF_OPPEDGE_REF
#undef CORNER_OF_OPPEDGE_TAG
#undef CORNER_OF_SIDE
#undef CORNER_OF_SIDE_INV
#undef CORNER_OF_SIDE_INV_REF
#undef CORNER_OF_SIDE_INV_TAG
#undef CORNER_OF_SIDE_PTR
#undef CORNER_OF_SIDE_REF
#undef CORNER_OF_SIDE_TAG
#undef CORNER_OPP_TO_SIDE
#undef CORNER_OPP_TO_SIDE_REF
#undef CORNER_OPP_TO_SIDE_TAG
#undef CORNERS_OF_BND_SEG
#undef CORNERS_OF_EDGE
#undef CORNERS_OF_ELEM
#undef CORNERS_OF_REF
#undef CORNERS_OF_SIDE
#undef CORNERS_OF_SIDE_REF
#undef CORNERS_OF_SIDE_TAG
#undef CORNERS_OF_TAG
#undef CORNERTYPE
#undef COUNTELEMENT
#undef COUNTNODE
#undef COUNTVECTOR
#undef COUNTVERTEX
#undef CP_EUKLIDNORM
#undef CP_LIMCOMB
#undef CPP_FRONTEND
#undef CP_SCALARPRODUCT
#undef CP_SUBTRACT
#undef CTRL
#undef CTRL2
#undef CURRENTLEVEL
#undef CVECT
#undef __CW__
#undef CW_FREE
#undef CW_INIT
#undef CW_INIT_UNUSED
#undef CW_READ
#undef CW_READ_STATIC
#undef CW_USED
#undef CW_WRITE
#undef CW_WRITE_STATIC
#undef DATAFORMAT
#undef __DDD__
#undef DDD_ATTR_NULL
#undef DDD_DOMAIN_DATA
#undef DDD_EXTRA_DATA
#undef DDD_GID_TO_INT
#undef DDD_IdentifyBegin
#undef DDD_IdentifyEnd
#undef DDD_IdentifyNumber
#undef DDD_IFAOneway
#undef DDD_InfoAttr
#undef DDD_InfoGlobalId
#undef DDD_InfoPriority
#undef DDD_InfoType
#undef DDD_OBJ
#undef DDD_PRIO_ENV
#undef DDD_PRIO_NULL
#undef DDD_PrioritySet
#undef DDD_PROC_NULL
#undef DDDTYPE
#undef DDD_TYPE_BY_HANDLER
#undef DDD_TYPE_NULL
#undef DDD_VERSION
#undef __DEBUG__
#undef _DEBUG_CW_
#undef DEBUG_MODE
#undef DEBUG_TIME
#undef DEBUG_TIME_INC
#undef DEBUG_TIME_MAX
#undef DEBUG_TIME_RESET
#undef DEC_NO_OF_ELEM
#undef DECNOOFNODE
#undef DECOUPLED
#undef DECOUPLED_LEN
#undef DECOUPLED_SHIFT
#undef DEFAULT_NAMES
#undef DEL_VECTOR
#undef DIAGMATRIXTYPE
#undef DIM
#undef DIMENSION_H
#undef DIM_MAX
#undef DIM_OF_BND
#undef DIM_OF_BND_MAX
#undef DIRSEP
#undef DMTP
#undef DNDS
#undef DNDT
#undef __DOMAIN__
#undef DOM_EVAL_SD_UNKNOWN
#undef DOM_LOC_X
#undef DOM_LOC_Y
#undef DOM_PARAM_OFFSET
#undef DOWNGRID
#undef DSIGNUM
#undef DUNE_UGINCLUDES_HH
#undef EATTR
#undef EBUILDCON
#undef EBUILDCON_LEN
#undef EBUILDCON_SHIFT
#undef ECLASS
#undef ECLASS_LEN
#undef ECLASS_SHIFT
#undef EDATA
#undef EDATA_DEF_IN_GRID
#undef EDATA_DEF_IN_MG
#undef EDDATA
#undef EDGENEW
#undef EDGENEW_LEN
#undef EDGENEW_SHIFT
#undef EDGE_OF_CORNER
#undef EDGE_OF_CORNER_REF
#undef EDGE_OF_CORNER_TAG
#undef EDGE_OFFSET
#undef EDGE_OF_SIDE
#undef EDGE_OF_SIDE_REF
#undef EDGE_OF_SIDE_TAG
#undef EDGE_OF_TWO_SIDES
#undef EDGE_OF_TWO_SIDES_REF
#undef EDGE_OF_TWO_SIDES_TAG
#undef EDGE_ON_BND
#undef EDGES_OF_CORNER
#undef EDGES_OF_CORNER_REF
#undef EDGES_OF_CORNER_TAG
#undef EDGES_OF_ELEM
#undef EDGES_OF_REF
#undef EDGES_OF_SIDE
#undef EDGES_OF_SIDE_REF
#undef EDGES_OF_SIDE_TAG
#undef EDGES_OF_TAG
#undef EDGE_WITH_CORNERS
#undef EDGE_WITH_CORNERS_REF
#undef EDGE_WITH_CORNERS_TAG
#undef EDID_FFMT
#undef EDID_FFMTE
#undef EDID_FFMTX
#undef EDID_FMT
#undef EDID_FMTE
#undef EDID_FMTX
#undef EDID_PRT
#undef EDID_PRTE
#undef EDID_PRTX
#undef EDSUBDOM
#undef EDSUBDOM_LEN
#undef EDSUBDOM_SHIFT
#undef EDVECTOR
#undef EFATHER
#undef EGHOST
#undef EGHOSTPRIO
#undef EGID
#undef EHGHOST
#undef EHGHOSTPRIO
#undef EIDCNT
#undef EID_FFMT
#undef EID_FFMTE
#undef EID_FFMTX
#undef EID_FMT
#undef EID_FMTE
#undef EID_FMTX
#undef EID_PRT
#undef EID_PRTE
#undef EID_PRTX
#undef ELEM_BNDS
#undef ELEMENTFMT
#undef ELEMENT_OFFSET
#undef ELEMENT_PTR
#undef __ELEMENTS__
#undef ELEMS_OF_NODE_MAX
#undef EMASTER
#undef EMASTERPRIO
#undef EMDD_EE
#undef EMDD_EE_PTR
#undef ENCOPIES
#undef ENDDEBUG
#undef ENDPAR
#undef END_UGDIM_NAMESPACE
#undef END_UG_NAMESPACE
#undef ENVDIR_DOWN
#undef ENVITEM_DOWN
#undef ENVITEM_LOCKED
#undef ENVITEM_NAME
#undef ENVITEM_TYPE
#undef EPRIO
#undef EPROCLIST
#undef EPROCPRIO
#undef ETA
#undef EVDD_E
#undef EVDD_E_PTR
#undef EVD_NCOMP
#undef EVECTOR
#undef EVEN
#undef EVGHOST
#undef EVGHOSTPRIO
#undef EVHGHOST
#undef EVHGHOSTPRIO
#undef __EVM__
#undef EVM_LOCKED
#undef __EXCHANGE_CONNECTIONS__
#undef EX_MAT
#undef EXTENSION_MAX
#undef FALSE
#undef FATHER_SIDE_OFFSET
#undef FINE_GRID_DOF
#undef FINE_GRID_DOF_LEN
#undef FINE_GRID_DOF_SHIFT
#undef FIRSTELEMENT
#undef FIRSTELEMSIDE
#undef FIRSTNODE
#undef FIRSTPART_OF_LIST
#undef FIRSTVECTOR
#undef FIRSTVERTEX
#undef FLAG
#undef FLAG_OFFSET
#undef FMT_CONN_DEPTH_MAX
#undef FMT_CONN_DEPTH_PTR
#undef FMT_CONN_DEPTH_TP
#undef FMT_ELEM_DATA
#undef FMT_MAX_PART
#undef FMT_MAX_TYPE
#undef FMT_N2T
#undef FMT_NB_DEPTH
#undef FMT_NODE_DATA
#undef FMT_NODE_ELEM_LIST
#undef FMT_PO2T
#undef FMT_PR_GRID
#undef FMT_PR_MAT
#undef FMT_PR_MG
#undef FMT_PR_VEC
#undef FMT_PR_VERTEX
#undef FMT_SET_N2T
#undef FMT_S_IMAT_TP
#undef FMT_S_MATPTR
#undef FMT_S_MAT_TP
#undef FMT_S_MG
#undef FMT_S_VEC_TP
#undef FMT_S_VERTEX
#undef FMT_T2N
#undef FMT_T2O
#undef FMT_T2P
#undef FMT_TYPE_IN_PART
#undef FMT_TYPE_USES_OBJ
#undef FMT_USES_OBJ
#undef FMT_VTYPE_NAME
#undef FORMAT
#undef __FORMATS__
#undef _FPTR
#undef FROM_VTNAME
#undef FSIGNUM
#undef FULLREFINELEVEL
#undef FULL_REFRULE
#undef FULL_REFRULE_0_5
#undef FULL_REFRULE_1_3
#undef FULL_REFRULE_2_4
#undef FULL_TPLT
#undef GATTR
#undef GBYTE
#undef __GENERAL__
#undef GENERAL_CW
#undef GENERAL_OFFSET
#undef GENERATED_NAMES_SEPERATOR
#undef GEN_MGUD
#undef GEN_MGUD_ADR
#undef GetAllSons
#undef GETGID
#undef GET_MATRIX
#undef GetMemoryForObject
#undef GetTmpMem
#undef GFIRSTBV
#undef GFORMAT
#undef GHOST
#undef GHOSTPRIO
#undef GID
#undef GIDFMT
#undef GLASTBV
#undef GLEVEL
#undef __GM__
#undef GM_TAKE_NONSKIP
#undef GM_TAKE_SKIP
#undef GRAPE_SUPPORT
#undef __GRAPE_TRUE__
#undef GREENCLASS
#undef GRID_ATTR
#undef GRID_OFFSET
#undef GRID_ON_LEVEL
#undef GRID_STATUS_OFFSET
#undef GSTATUS
#undef HAS_DDDHDR
#undef HDR
#undef HDRELEMENT
#undef HDRNODE
#undef HDRVECTOR
#undef HDRVERTEX
#undef HEAPCHECK
#undef HEAPFAULT
#undef __HEAPS__
#undef HGHOST
#undef HGHOSTPRIO
#undef HiWrd
#undef ID
#undef IDENT_ONLY_NEW
#undef ID_FFMT
#undef ID_FFMTE
#undef ID_FFMTX
#undef ID_FMT
#undef ID_FMTE
#undef ID_FMTX
#undef ID_PRT
#undef ID_PRTE
#undef ID_PRTX
#undef IFDEBUG
#undef INC_NO_OF_ELEM
#undef INCNOOFNODE
#undef INIT
#undef __INITUG__
#undef INNER_BOUNDARY
#undef INNER_SIDE
#undef INNER_SIZE_TAG
#undef __INTERPOLATION_MATRIX__
#undef INVERSE_TRANSFORMATION
#undef IS_BLOCK_DEFINED
#undef IS_ENVDIR
#undef IS_MGUDBLOCK_DEF
#undef ISNaN
#undef IS_REFINED
#undef KBYTE
#undef KEEP_VECTOR
#undef LASTELEMENT
#undef LASTNODE
#undef LASTPART_OF_LIST
#undef LASTVECTOR
#undef LASTVERTEX
#undef LCVECT
#undef LDATA
#undef LEAFELEM
#undef LELEM
#undef LEVEL
#undef LEVEL_LEN
#undef LEVEL_SHIFT
#undef LINK
#undef LINK0
#undef LINK1
#undef LINK_OFFSET
#undef LINKX
#undef LIOBJ
#undef LISTPART2PRIO
#undef LISTPART_FIRSTELEMENT
#undef LISTPART_FIRSTNODE
#undef LISTPART_FIRSTVECTOR
#undef LISTPART_FIRSTVERTEX
#undef LISTPART_LASTELEMENT
#undef LISTPART_LASTNODE
#undef LISTPART_LASTVECTOR
#undef LISTPART_LASTVERTEX
#undef LOCAL_COORD_OF_ELEM
#undef LOCAL_COORD_OF_REF
#undef LOCAL_COORD_OF_TAG
#undef LOCAL_TO_GLOBAL
#undef LOCAL_TO_GLOBAL_2D
#undef LOCAL_TO_GLOBAL_3D
#undef LOCAL_TO_GLOBAL_BND
#undef LOCAL_TO_GLOBAL_HEXAHEDRON
#undef LOCAL_TO_GLOBAL_PRISM
#undef LOCAL_TO_GLOBAL_PYRAMID
#undef LOCAL_TO_GLOBAL_QUADRILATERAL
#undef LOCAL_TO_GLOBAL_TETRAHEDRON
#undef LOCAL_TO_GLOBAL_TRIANGLE
#undef LOFFSET
#undef LOFFSET_LEN
#undef LOFFSET_SHIFT
#undef LoWrd
#undef LV_MOD_DEFAULT
#undef M1_INVERT
#undef M2_ADD
#undef M2_COPY
#undef M2_DET
#undef M2_INVERT
#undef M2_INVERT_STD
#undef M2_LINCOMB
#undef M2_MAXNORM
#undef M2_SCALE
#undef M2_TIMES_M2
#undef M2_TIMES_V2
#undef M3_ADD
#undef M3_ADDMATRIX
#undef M3_COPY
#undef M3_DET
#undef M3_INVERT
#undef M3_LINCOMB
#undef M3_SCALE
#undef M3_TIMES_M3
#undef M3_TIMES_V3
#undef M4_CLEAR
#undef M4_COPY
#undef M4_TIMES_M4
#undef MACTIVE
#undef MACTIVE_LEN
#undef MACTIVE_SHIFT
#undef MADJ
#undef MAPPED_BND_OBJT_TAG
#undef MAPPED_INNER_OBJT_TAG
#undef MARK
#undef MARK2PAT
#undef MARK2PATTERN
#undef MARK2RULE
#undef MARK2RULEADR
#undef MARKCLASS
#undef MARKCLASS_LEN
#undef MARKCLASS_SHIFT
#undef MARKCOUNT
#undef MARKED
#undef MARK_LEN
#undef MARK_OF_RULE
#undef MARK_SHIFT
#undef MARK_STACK_SIZE
#undef MarkTmpMem
#undef MASTER
#undef MASTERPRIO
#undef MATELEM
#undef MATRIX_OFFSET
#undef MATRIXTYPE
#undef MAX
#undef MAX_BV_LEVEL
#undef MAX_BV_NUMBER
#undef MAXCONNECTIONS
#undef MAX_CONTROL_ENTRIES
#undef MAX_CONTROL_WORDS
#undef MAX_CORNERS_OF_EDGE
#undef MAX_CORNERS_OF_ELEM
#undef MAX_CORNERS_OF_ELEM_2D
#undef MAX_CORNERS_OF_ELEM_3D
#undef MAX_CORNERS_OF_ELEM_DIM
#undef MAX_CORNERS_OF_SIDE
#undef MAXDDDTYPES
#undef MAXDOMPARTS
#undef MAX_EDGES_OF_CORNER
#undef MAX_EDGES_OF_ELEM
#undef MAX_EDGES_OF_ELEM_2D
#undef MAX_EDGES_OF_ELEM_3D
#undef MAX_EDGES_OF_ELEM_DIM
#undef MAX_EDGES_OF_SIDE
#undef MAX_ELEM_VECTORS
#undef MAXFREEOBJECTS
#undef MAXLEVEL
#undef MAX_MAT_COMP
#undef MAX_MAT_COMP_TOTAL
#undef MAXMATRICES
#undef MAXNBLOCKS
#undef MAX_NDOF
#undef MAX_NDOF_MOD_32
#undef MAX_NEW_CORNERS
#undef MAX_NEW_CORNERS_2D
#undef MAX_NEW_CORNERS_3D
#undef MAX_NEW_CORNERS_DIM
#undef MAX_NEW_EDGES
#undef MAX_NEW_EDGES_2D
#undef MAX_NEW_EDGES_3D
#undef MAX_NEW_EDGES_DIM
#undef MAXOBJECTS
#undef MAX_PAR_DIST
#undef MAX_PATH_DEPTH
#undef MAX_PERIODIC_OBJ
#undef MAX_REFINED_CORNERS_DIM
#undef MAX_RULES
#undef MAXSELECTION
#undef MAX_SIDE_NODES
#undef MAX_SIDES_OF_EDGE
#undef MAX_SIDES_OF_ELEM
#undef MAX_SIDES_OF_ELEM_2D
#undef MAX_SIDES_OF_ELEM_3D
#undef MAX_SIDES_OF_ELEM_DIM
#undef MAX_SIDES_TOUCHING
#undef MAX_SINGLE_MAT_COMP
#undef MAX_SINGLE_VEC_COMP
#undef MAX_SON_EDGES
#undef MAX_SONS_2D
#undef MAX_SONS_3D
#undef MAX_SONS_DIM
#undef MAX_SUB
#undef MAX_VEC_COMP
#undef MAXVECTORS
#undef MAXVOBJECTS
#undef MAXVTNAMES
#undef M_BDIM_INVERT
#undef MBYTE
#undef MCMP
#undef MCMP_I
#undef MCMP_J
#undef M_COMP_NAMES
#undef MD2_TIMES_V2
#undef MD3_TIMES_V3
#undef MD_COL_DATA_TYPES
#undef MD_COL_OBJ_USED
#undef MD_COLPTR
#undef MD_COLS_IN_MTYPE
#undef MD_cols_in_ro_co
#undef MD_COLS_IN_RT_CT
#undef MDEC
#undef MDEST
#undef MDESTINDEX
#undef MDESTTYPE
#undef MDESTTYPE_LEN
#undef MDESTTYPE_SHIFT
#undef MDIAG
#undef MDIAG_LEN
#undef MDIAG_SHIFT
#undef MD_IJ_CMP_OF_MTYPE
#undef MD_IJ_CMP_OF_RT_CT
#undef M_DIM_ADD
#undef M_DIM_COPY
#undef M_DIM_DET
#undef M_DIM_INVERT
#undef M_DIM_SCALE
#undef MD_ISDEF_IN_MTYPE
#undef MD_ISDEF_IN_RT_CT
#undef MD_IS_SCALAR
#undef MD_IS_SPARSE
#undef MD_MCMP_OF_MTYPE
#undef MD_mcmp_of_ro_co
#undef MD_MCMP_OF_RT_CT
#undef MD_MCMPPTR
#undef MD_MCMPPTR_OF_MTYPE
#undef MD_mcmpptr_of_ro_co
#undef MD_mcmpptr_of_ro_co_mod
#undef MD_MCMPPTR_OF_RT_CT
#undef MD_MG
#undef MD_MTYPE_OFFSET
#undef MD_NCMPS_IN_MTYPE
#undef MD_NCMPS_IN_RT_CT
#undef MD_nr_nc_mcmpptr_of_ro_co
#undef MD_OFFSETPTR
#undef MDOWN
#undef MDOWN_LEN
#undef MDOWN_SHIFT
#undef MD_ROW_DATA_TYPES
#undef MD_ROW_OBJ_USED
#undef MD_ROWPTR
#undef MD_rows_cols_in_ro_co
#undef MD_ROWS_IN_MTYPE
#undef MD_rows_in_ro_co
#undef MD_ROWS_IN_RT_CT
#undef MD_RT_CT_OFFSET
#undef MD_SCALCMP
#undef MD_SCAL_CTYPEMASK
#undef MD_SCAL_RTYPEMASK
#undef MD_SM
#undef MD_SMP
#undef MD_SUCC_COMP
#undef MD_TIMES_V_DIM
#undef MGBNDSEGDESC
#undef MG_BVP
#undef MG_BVPD
#undef MG_COARSE_FIXED
#undef MG_FILENAME
#undef MGFORMAT
#undef MG_GENPURP
#undef MGHEAP
#undef MG_MAGIC_COOKIE
#undef MG_MARK_KEY
#undef MGNAME
#undef MGNDELEMBLK
#undef MGNDELEMBLKENTRY
#undef MGNDELEMOFFS
#undef MGNDELEMPTRARRAY
#undef MGNOOFCORNERS
#undef MG_NPROPERTY
#undef MG_SAVED
#undef MGSTATUS
#undef MG_USER_HEAP
#undef MGVERTEX
#undef MIDNODE
#undef MIDTYPE
#undef MIN
#undef MINC
#undef MIN_HEAP_SIZE
#undef __MISC__
#undef MLOWER
#undef MLOWER_LEN
#undef MLOWER_SHIFT
#undef MM2_COPY
#undef MM2_DET
#undef MM2_TIMES_V2
#undef MM3_COPY
#undef MM3_DET
#undef MM3_TIMES_V3
#undef MM_DIM_COPY
#undef MM_DIM_DET
#undef MM_TIMES_V_DIM
#undef MMYCON
#undef MNEW
#undef MNEW_LEN
#undef MNEW_SHIFT
#undef MNEXT
#undef MODEL
#undef MODIFIED
#undef MODIFIED_LEN
#undef MODIFIED_SHIFT
#undef MOFFSET
#undef MOFFSET_LEN
#undef MOFFSET_SHIFT
#undef MOVE
#undef MOVED
#undef MOVED_LEN
#undef MOVED_SHIFT
#undef MOVE_LEN
#undef MOVE_SHIFT
#undef MROOT
#undef MROOTTYPE
#undef MROOTTYPE_LEN
#undef MROOTTYPE_SHIFT
#undef MSIZE_LEN
#undef MSIZEMAX
#undef MSIZE_SHIFT
#undef MSTRONG
#undef MT1_TIMES_V1
#undef MT2_TIMES_V2
#undef MT3_TIMES_V3
#undef MT_CCOMP
#undef MT_CCOMPS
#undef MT_COMP
#undef MT_COMPNAME
#undef MT_COMPNAMES
#undef MT_COMPS
#undef M_TIMES_V_DIM
#undef MT_MCMP_OF_MTYPE
#undef MT_MCMPPTR
#undef MT_MCMPPTR_OF_MTYPE
#undef MT_NSUB
#undef MTP
#undef MT_RCOMP
#undef MT_RCOMPS
#undef MT_SM
#undef MT_SMP
#undef MT_SUB
#undef MT_TIMES_V_BDIM
#undef MT_TIMES_V_DIM
#undef MTYPE
#undef MTYPE_CT
#undef MTYPE_RT
#undef MULTIGRID_STATUS_OFFSET
#undef MUP
#undef MUP_LEN
#undef MUPPER
#undef MUPPER_LEN
#undef MUPPER_SHIFT
#undef MUP_SHIFT
#undef MUSED
#undef MVALUE
#undef MVALUEPTR
#undef MYEDGE
#undef MYMG
#undef MYVERTEX
#undef NAMELENSTR
#undef NBELEM
#undef NBNODE
#undef NC
#undef NCLASS
#undef NCLASS_LEN
#undef NCLASS_SHIFT
#undef NCOPIES
#undef NDATA
#undef NDATA_DEF_IN_GRID
#undef NDATA_DEF_IN_MG
#undef NDELEM_BLKS_MAX
#undef NE
#undef NELIST_DEF_IN_GRID
#undef NELIST_DEF_IN_MG
#undef NETGEN_SUPPORT
#undef NEW_DEFECT
#undef NEW_DEFECT_LEN
#undef NEW_DEFECT_SHIFT
#undef NEW_EDIDENT
#undef NEW_EDIDENT_LEN
#undef NEWEL
#undef NEWEL_LEN
#undef NEWEL_SHIFT
#undef NEW_NIDENT
#undef NEW_NIDENT_LEN
#undef NEXT
#undef NEXT_ENVITEM
#undef NEXTSIDE
#undef NEXTSIDEHEX
#undef NEXTSIDEMASK
#undef NEXTSIDEMASKHEX
#undef NFATHER
#undef NFATHEREDGE
#undef NIDCNT
#undef NIMAT
#undef NMATOFFSETS
#undef NMATTYPES
#undef NMATTYPES_NORMAL
#undef NN
#undef NNCLASS
#undef NNCLASS_LEN
#undef NNCLASS_SHIFT
#undef NN_PRIO
#undef NO
#undef NO_BLOCKVECTOR
#undef NO_CENTER_NODE
#undef NOCLASS
#undef NODE_ELEMENT_LIST
#undef NODEFMT
#undef NO_DELETE_OVERLAP2
#undef NO_DELETE_OVERLAP2_LEN
#undef NODE_OFFSET
#undef NODE_OF_RULE
#undef NODEORD
#undef NODEORD_LEN
#undef NODEORD_SHIFT
#undef NO_IDENT
#undef NO_MSGID
#undef NO_NODES_OF_BLK
#undef NO_OF_ELEM
#undef NO_OF_ELEM_LEN
#undef NO_OF_ELEM_MAX
#undef NO_OF_ELEM_SHIFT
#undef NOOFNODE
#undef NOOFNODE_LEN
#undef NOOFNODEMAX
#undef NOOFNODE_SHIFT
#undef NO_OPTION_CHECK
#undef NOTUSED
#undef NOVTYPE
#undef NPROP
#undef NPROP_LEN
#undef NPROP_SHIFT
#undef NS
#undef NS_DIM_PREFIX
#undef NSONS
#undef NSONS_LEN
#undef NSONS_OF_RULE
#undef NSONS_SHIFT
#undef NS_PREFIX
#undef NSUBDOM
#undef NSUBDOM_LEN
#undef NSUBDOM_SHIFT
#undef NT
#undef NT_PRIO
#undef NTYPE
#undef NTYPE_LEN
#undef NTYPE_SHIFT
#undef NU
#undef NV
#undef NVEC
#undef NVECOFFSETS
#undef NVEC_PRIO
#undef NVECTOR
#undef NVECTYPES
#undef NV_PRIO
#undef OBJ_LEN
#undef _OBJREF
#undef OBJ_SHIFT
#undef OBJT
#undef OBJT_MAX
#undef ODD
#undef OFF
#undef OFFSET_IN_HEAP
#undef OFFSET_IN_MGUD
#undef ON
#undef ONEDGE
#undef ONEDGE_LEN
#undef ONEDGE_SHIFT
#undef ONNBSIDE
#undef ONNBSIDE_LEN
#undef ONNBSIDE_SHIFT
#undef ONSIDE
#undef ONSIDE_LEN
#undef ONSIDE_SHIFT
#undef OPPOSITE_EDGE
#undef OPPOSITE_EDGE_REF
#undef OPPOSITE_EDGE_TAG
#undef ORD_CYC
#undef ORD_GEN
#undef ORD_LIN
#undef OUT_OF_RANGE
#undef __OVERLAP2__
#undef PAR
#undef __PARALLEL_H__
#undef __PARGM_H__
#undef PARHDR
#undef PARHDRE
#undef PARHDRV
#undef PARTITION
#undef PATHDEPTH
#undef PATHDEPTHMASK
#undef PATHDEPTHSHIFT
#undef PAT_OF_RULE
#undef PATTERN
#undef PATTERN2MARK
#undef PATTERN2RULE
#undef PATTERN_LEN
#undef PATTERN_OF_RULE
#undef PATTERN_SHIFT
#undef __PERIODIC_BOUNDARY__
#undef PFIRSTELEMENT
#undef PFIRSTNODE
#undef PFIRSTVECTOR
#undef PFIRSTVERTEX
#undef PFMT
#undef PI
#undef PLASTELEMENT
#undef PLASTNODE
#undef PLASTVECTOR
#undef PLASTVERTEX
#undef POW
#undef POW2
#undef __PPIF__
#undef PPIF_FAILURE
#undef PPIF_FRONTEND_C
#undef PPIF_NAMESPACE_H
#undef PPIF_NS_PREFIX
#undef PPIF_SUCCESS
#undef PPREDVC
#undef PREDE
#undef PREDMAX
#undef PREDN
#undef PREDNEW0
#undef PREDNEW1
#undef PREDNEW2
#undef PREDV
#undef PREDVC
#undef PREV_ENVITEM
#undef PrintDebug
#undef PRINTDEBUG
#undef PRINTDEBUG_EXT
#undef PRINT_LIST
#undef PRIO
#undef PRIO2INDEX
#undef PRIO2LISTPART
#undef PRIOCOUNTELEMENT
#undef PRIOCOUNTNODE
#undef PRIOCOUNTVECTOR
#undef PRIOCOUNTVERTEX
#undef PRIO_FIRSTELEMENT
#undef PRIO_FIRSTNODE
#undef PRIO_FIRSTVECTOR
#undef PRIO_FIRSTVERTEX
#undef PRIO_LASTELEMENT
#undef PRIO_LASTNODE
#undef PRIO_LASTVECTOR
#undef PRIO_LASTVERTEX
#undef PROCLIST
#undef PROCPRIO
#undef PROP
#undef PROPERTY_OFFSET
#undef PROP_LEN
#undef PROP_SHIFT
#undef PutFreeObject
#undef PVCOUNT
#undef QUOT
#undef RAND_MSG_SIZE
#undef RCSID
#undef RCS_ID
#undef RCSID1
#undef RCS_ID1
#undef RCSIDAUX
#undef RCS_ID_AUX
#undef RCSIDAUX1
#undef RCS_ID_AUX1
#undef READ_FLAG
#undef READ_VEC_COLLECT_FLAG
#undef READ_VEC_CONS_FLAG
#undef REAL
#undef REDCLASS
#undef REF2TAG
#undef REFINE
#undef __REFINE__
#undef REFINECLASS
#undef REFINECLASS_LEN
#undef REFINECLASS_SHIFT
#undef REFINE_ELEMENT_LIST
#undef REFINE_GRID_LIST
#undef REFINEINFO
#undef REFINE_LEN
#undef REFINE_MULTIGRID_LIST
#undef REFINE_SHIFT
#undef REFINESTEP
#undef ReleaseTmpMem
#undef REP_ERR_ENCOUNTERED
#undef REP_ERR_FILE
#undef REP_ERR_GOTO
#undef REP_ERR_INC
#undef REP_ERR_MAX
#undef REP_ERR_RESET
#undef REP_ERR_RETURN
#undef REP_ERR_RETURN_PTR
#undef REP_ERR_RETURN_VOID
#undef RESETGSTATUS
#undef RESETMGSTATUS
#undef RETURN
#undef REVERSE
#undef RINFO_MAX
#undef ROOT_DIR
#undef RULE2MARK
#undef RULE2PAT
#undef RULE2PATTERN
#undef __RULEMANAGER__
#undef SELECTIONMODE
#undef SELECTIONOBJECT
#undef SELECTIONSIZE
#undef SETADDPATTERN
#undef SETAUXEDGE
#undef SET_BNDS
#undef SETBVDOWNTYPE
#undef SETBV_GC
#undef SETBVLEVEL
#undef SETBVORIENTATION
#undef SETBVTVTYPE
#undef SET_CENTERNODE
#undef SETCEXTRA
#undef SETCOARSEN
#undef SET_CORNER
#undef SETCUSED
#undef SETDECOUPLED
#undef SETEBUILDCON
#undef SETECLASS
#undef SET_EDATA
#undef SETEDGENEW
#undef SETEDSUBDOM
#undef SET_EFATHER
#undef SETEPRIO
#undef SETEPRIOX
#undef SET_EVECTOR
#undef SETFINE_GRID_DOF
#undef SET_FLAG
#undef SETGLOBALGSTATUS
#undef SETHEAPFAULT
#undef SetHiWrd
#undef SET_LELEM
#undef SETLEVEL
#undef SETLOFFSET
#undef SetLoWrd
#undef SETMACTIVE
#undef SETMARK
#undef SETMARKCLASS
#undef SETMARKCOUNT
#undef SETMDESTTYPE
#undef SETMDIAG
#undef SETMDOWN
#undef SETMLOWER
#undef SETMNEW
#undef SETMODIFIED
#undef SETMOFFSET
#undef SETMOVE
#undef SETMOVED
#undef SETMROOTTYPE
#undef SETMSIZE
#undef SETMUP
#undef SETMUPPER
#undef SETMUSED
#undef SET_NBELEM
#undef SETNCLASS
#undef SETNEW_DEFECT
#undef SETNEW_EDIDENT
#undef SETNEWEL
#undef SETNEW_NIDENT
#undef SETNEXTSIDE
#undef SETNFATHER
#undef SETNFATHEREDGE
#undef SETNNCLASS
#undef SETNO_DELETE_OVERLAP2
#undef SETNODEORD
#undef SET_NO_OF_ELEM
#undef SETNOOFNODE
#undef SETNPROP
#undef SETNSONS
#undef SETNSUBDOM
#undef SETNTYPE
#undef SETOBJT
#undef SETONEDGE
#undef SETONNBSIDE
#undef SETONSIDE
#undef SET_ORD_GCL
#undef SETPATHDEPTH
#undef SETPATTERN
#undef SETPREDMAX
#undef SETPREDNEW0
#undef SETPREDNEW1
#undef SETPREDNEW2
#undef SETPRIO
#undef SETPRIOX
#undef SETPROP
#undef SETPVCOUNT
#undef SETREAL
#undef SETREFINE
#undef SETREFINECLASS
#undef SETREFINESTEP
#undef SETSIDEPATTERN
#undef SET_SON
#undef SETSUBDOMAIN
#undef SET_SVECTOR
#undef SETTAG
#undef SETTHEFLAG
#undef SETUPDATE_GREEN
#undef SETUSED
#undef SETVACTIVE
#undef SETVBUILDCON
#undef SETVCCOARSE
#undef SETVCCUT
#undef SETVCFLAG
#undef SETVCLASS
#undef SETVCOUNT
#undef SETVCUSED
#undef SETVDATATYPE
#undef SETVDOWN
#undef SET_VEC_COLLECT__FLAG
#undef SET_VEC_CONS__FLAG
#undef SETVECSKIPBIT
#undef SETVECTORSIDE
#undef SETVNCLASS
#undef SETVNEW
#undef SETVOTYPE
#undef SETVPART
#undef SETVSKIPME
#undef SETVTYPE
#undef SETVUP
#undef SETVXPRIO
#undef SETVXPRIOX
#undef SETXFERMATX
#undef SETXFERVECTOR
#undef SFIRSTNODE
#undef SFIRSTVECTOR
#undef SFIRSTVERTEX
#undef __SHAPES__
#undef SideBndCond
#undef SIDE_NORMAL
#undef SIDE_NORMAL_2D
#undef SIDE_NORMAL_3D
#undef SIDE_ON_BND
#undef SIDE_OPP_TO_CORNER
#undef SIDE_OPP_TO_CORNER_REF
#undef SIDE_OPP_TO_CORNER_TAG
#undef SIDEPATTERN
#undef SIDEPATTERN_LEN
#undef SIDEPATTERN_SHIFT
#undef SIDES_OF_ELEM
#undef SIDES_OF_REF
#undef SIDES_OF_TAG
#undef SIDETYPE
#undef SIDE_WITH_EDGE
#undef SIDE_WITH_EDGE_REF
#undef SIDE_WITH_EDGE_TAG
#undef SIGNIFICANT_DIGITS
#undef SIGNUM
#undef SIZEOF_VHM
#undef SIZE_UNKNOWN
#undef __SM__
#undef SON
#undef SON_CORNER
#undef SON_CORNER_OF_RULE
#undef SON_NB
#undef SON_NB_OF_RULE
#undef SONNODE
#undef SONNODE_OF_NODE_OF_RULE
#undef SON_OF_NODE_OF_RULE
#undef SON_OF_RULE
#undef SON_PATH
#undef SON_PATH_OF_RULE
#undef SONS_OF_ELEM
#undef SONS_OF_TAG
#undef SON_TAG
#undef SON_TAG_OF_RULE
#undef SPID_BACK
#undef SPID_FORTH
#undef SPID_MD
#undef SPID_MDI
#undef SPID_NMD
#undef SPID_NMD_MAX
#undef SPID_NVD
#undef SPID_NVD_MAX
#undef SPID_VD
#undef SPID_VDI
#undef SQRT
#undef START
#undef START_UGDIM_NAMESPACE
#undef START_UG_NAMESPACE
#undef StaticControlWord
#undef StaticControlWordMask
#undef STAT_OUT
#undef __STD_DOMAIN__
#undef STR
#undef STR_BREAK
#undef STR_CASE
#undef STR_DEFAULT
#undef STR_SWITCH
#undef STR_SWITCH_END
#undef SUBDOMAIN
#undef SUBDOMAIN_LEN
#undef SUBDOMAIN_SHIFT
#undef SUBM_CCOMP
#undef SUBM_CCOMPS
#undef SUBM_COMP
#undef SUBM_COMPS
#undef SUBM_MCMP_OF_MTYPE
#undef SUBM_MCMPPTR
#undef SUBM_MCMPPTR_OF_MTYPE
#undef SUBM_NAME
#undef SUBM_RCOMP
#undef SUBM_RCOMPS
#undef SUBV_COMP
#undef SUBV_NAME
#undef SUBV_NCOMP
#undef SUBV_NCOMPS
#undef SUCCE
#undef SUCCN
#undef SUCCV
#undef SUCCVC
#undef SVECTOR
#undef SWAP
#undef SWITCHCLASS
#undef SYNC_ALL
#undef SYNC_CONTEXT
#undef SYNC_END
#undef TAG
#undef TAG_LEN
#undef TAG_OF_RULE
#undef TAGS
#undef TAG_SHIFT
#undef TET_COPY
#undef TET_RED
#undef TET_RED_0_5
#undef TET_RED_1_3
#undef TET_RED_2_4
#undef TET_RED_HEX
#undef THEFLAG
#undef THEFLAG_LEN
#undef THEFLAG_SHIFT
#undef __THREEDIM__
#undef TOPLEVEL
#undef TOPNODE
#undef TOTUSED_IN_HEAP
#undef TO_VTNAME
#undef TRANSFORMATION
#undef TRANSFORMATION_2D
#undef TRANSFORMATION_3D
#undef TRANSFORMATION_BND
#undef TRANSFORMATION_OF_HEXAHEDRON
#undef TRANSFORMATION_OF_PRISM
#undef TRANSFORMATION_OF_PYRAMID
#undef TRANSFORMATION_OF_QUADRILATERAL
#undef TRANSFORMATION_OF_TETRAHEDRON
#undef TRANSFORMATION_OF_TRIANGLE
#undef TRUE
#undef __TWODIM__
#undef __UDM__
#undef __UG__
#undef __UGENV__
#undef UG_GlobalMaxDOUBLE
#undef UG_GlobalMaxINT
#undef UG_GlobalMaxNDOUBLE
#undef UG_GlobalMaxNINT
#undef UG_GlobalMinDOUBLE
#undef UG_GlobalMinINT
#undef UG_GlobalMinNDOUBLE
#undef UG_GlobalMinNINT
#undef UG_GlobalSumDOUBLE
#undef UG_GlobalSumINT
#undef UG_GlobalSumNDOUBLE
#undef UG_GlobalSumNINT
#undef __UGM__
#undef UG_MSIZE
#undef UG_NAMESPACE_H
#undef UGRCSSTRING
#undef UG_RCS_STRING
#undef UGRCSSTRINGAUX
#undef UGTYPE
#undef UGTYPES_H
#undef UG_VERSION
#undef UNLINK
#undef UPDATE_GREEN
#undef UPDATE_GREEN_LEN
#undef UPDATE_GREEN_SHIFT
#undef UPGRID
#undef USED
#undef USED_LEN
#undef USED_SHIFT
#undef USE_FAMG
#undef USING_PPIF_NAMESPACE
#undef USING_UGDIM_NAMESPACE
#undef USING_UG_NAMESPACE
#undef USING_UG_NAMESPACES
#undef V1_ADD
#undef V1_ADD1
#undef V1_AVG2
#undef V1_AVG3
#undef V1_AVG4
#undef V1_CLEAR
#undef V1_COMPARE
#undef V1_COPY
#undef V1_EUKLIDNORM
#undef V1_EUKLIDNORM_OF_DIFF
#undef V1_ISEQUAL
#undef V1_ISZERO
#undef V1_LINCOMB
#undef V1_NORMAL
#undef V1_SCALAR_PRODUCT
#undef V1_SCALE
#undef V1_SCALEADD1
#undef V1_SCALESET
#undef V1_SCAL_PROD
#undef V1_SET
#undef V1_SUBTRACT
#undef V1_SUP
#undef V1_VECTOR_PRODUCT
#undef V2_ADD
#undef V2_ADD1
#undef V2_AVG2
#undef V2_AVG3
#undef V2_AVG4
#undef V2_CLEAR
#undef V2_COMPARE
#undef V2_COPY
#undef V2_EUKLIDNORM
#undef V2_EUKLIDNORM_OF_DIFF
#undef V2_ISEQUAL
#undef V2_ISZERO
#undef V2_LINCOMB
#undef V2_NORMAL
#undef V2_SCALAR_PRODUCT
#undef V2_SCALE
#undef V2_SCALEADD1
#undef V2_SCALESET
#undef V2_SCAL_PROD
#undef V2_SET
#undef V2_SUBTRACT
#undef V2_SUP
#undef V2_VECTOR_PRODUCT
#undef V3_ADD
#undef V3_ADD1
#undef V3_AVG2
#undef V3_AVG3
#undef V3_AVG4
#undef V3_CLEAR
#undef V3_COMPARE
#undef V3_COPY
#undef V3_EUKLIDNORM
#undef V3_EUKLIDNORM_OF_DIFF
#undef V3_ISEQUAL
#undef V3_ISZERO
#undef V3_LINCOMB
#undef V3_SCALAR_PRODUCT
#undef V3_SCALE
#undef V3_SCALEADD1
#undef V3_SCALESET
#undef V3_SCAL_PROD
#undef V3_SET
#undef V3_SUBTRACT
#undef V3_SUP
#undef V3_VECTOR_PRODUCT
#undef VACTIVE
#undef VACTIVE_LEN
#undef VACTIVE_SHIFT
#undef V_BDIM_ADD
#undef V_BDIM_ADD1
#undef V_BDIM_AVG2
#undef V_BDIM_AVG3
#undef V_BDIM_AVG4
#undef V_BDIM_CLEAR
#undef V_BDIM_COMPARE
#undef V_BDIM_COPY
#undef V_BDIM_EUKLIDNORM
#undef V_BDIM_EUKLIDNORM_OF_DIFF
#undef V_BDIM_ISEQUAL
#undef V_BDIM_ISZERO
#undef V_BDIM_LINCOMB
#undef V_BDIM_Normalize
#undef V_BDIM_SCALAR_PRODUCT
#undef V_BDIM_SCALE
#undef V_BDIM_SCALEADD1
#undef V_BDIM_SCALESET
#undef V_BDIM_SCAL_PROD
#undef V_BDIM_SET
#undef V_BDIM_SUBTRACT
#undef V_BDIM_SUP
#undef V_BDIM_VECTOR_PRODUCT
#undef V_BNDP
#undef VBUILDCON
#undef VBVD
#undef VCCOARSE
#undef VCCOARSE_LEN
#undef VCCOARSE_SHIFT
#undef VCCUT
#undef VCCUT_LEN
#undef VCCUT_SHIFT
#undef VCFLAG
#undef VCLASS
#undef VCLASS_LEN
#undef VCLASS_SHIFT
#undef V_COMP_NAMES
#undef VCOUNT
#undef VCOUNT_LEN
#undef VCOUNT_SHIFT
#undef VCUSED
#undef VDATA
#undef VDATATYPE
#undef VDATATYPE_LEN
#undef VDATATYPE_SHIFT
#undef VD_cmp_of_otype
#undef VD_CMP_OF_TYPE
#undef VD_cmpptr_of_otype
#undef VD_cmpptr_of_otype_mod
#undef VD_CMPPTR_OF_TYPE
#undef VD_DATA_TYPES
#undef VD_IDENT
#undef VD_IDENT_PTR
#undef V_DIM_ADD
#undef V_DIM_ADD1
#undef V_DIM_AVG2
#undef V_DIM_AVG3
#undef V_DIM_AVG4
#undef V_DIM_CLEAR
#undef V_DIM_COMPARE
#undef V_DIM_COPY
#undef V_DIM_EUKLIDNORM
#undef V_DIM_EUKLIDNORM_OF_DIFF
#undef V_DIM_ISEQUAL
#undef V_DIM_ISZERO
#undef V_DIM_LINCOMB
#undef V_DIM_NORMAL
#undef V_DIM_Normalize
#undef V_DIM_SCALAR_PRODUCT
#undef V_DIM_SCALE
#undef V_DIM_SCALEADD1
#undef V_DIM_SCALESET
#undef V_DIM_SCAL_PROD
#undef V_DIM_SET
#undef V_DIM_SP
#undef V_DIM_SUBTRACT
#undef V_DIM_SUP
#undef V_DIM_VECTOR_PRODUCT
#undef VD_ISDEF_IN_TYPE
#undef VD_IS_SCALAR
#undef VD_MAXTYPE
#undef VD_MG
#undef VD_MINTYPE
#undef VD_ncmp_cmpptr_of_otype
#undef VD_NCMPPTR
#undef VD_ncmps_in_otype
#undef VD_NCMPS_IN_TYPE
#undef VD_NCOMP
#undef VD_NID
#undef VD_OBJ_USED
#undef VD_OFFSET
#undef VD_OFFSETPTR
#undef VDOWN
#undef VD_SCALCMP
#undef VD_SCALTYPEMASK
#undef VD_SUCC_COMP
#undef VEC_DEF_IN_OBJ_OF_GRID
#undef VEC_DEF_IN_OBJ_OF_MG
#undef VECSKIP
#undef VECSKIPBIT
#undef VECTORFMT
#undef VECTOR_OFFSET
#undef VECTORSIDE
#undef VECTORSIDE_LEN
#undef VECTORSIDE_SHIFT
#undef Vertex_BndCond
#undef VERTEXFMT
#undef VERTEX_OFFSET
#undef VFATHER
#undef VFULLSKIP
#undef VGHOST
#undef VGHOSTPRIO
#undef VGID
#undef VHGHOST
#undef VHGHOSTPRIO
#undef VIDCNT
#undef VID_FFMT
#undef VID_FFMTE
#undef VID_FFMTX
#undef VID_FMT
#undef VID_FMTE
#undef VID_FMTX
#undef VID_PRT
#undef VID_PRTE
#undef VID_PRTX
#undef V_IN_DATATYPE
#undef VINDEX
#undef VINDEX_FFMT
#undef VINDEX_FFMTE
#undef VINDEX_FFMTX
#undef VINDEX_FMT
#undef VINDEX_FMTE
#undef VINDEX_FMTX
#undef VINDEX_PRT
#undef VINDEX_PRTE
#undef VINDEX_PRTX
#undef VISTART
#undef VMATCH
#undef VM_COMP_NAME
#undef VM_COMP_NAMEPTR
#undef VM_COMPPTR
#undef VM_LOCKED
#undef VMYEDGE
#undef VMYELEMENT
#undef VMYNODE
#undef VNCLASS
#undef VNCLASS_LEN
#undef VNCLASS_SHIFT
#undef VNEW
#undef VNEW_LEN
#undef VNEW_SHIFT
#undef VOBJECT
#undef VOID_NBELEM
#undef VOTYPE
#undef VOTYPE_LEN
#undef VOTYPE_SHIFT
#undef VPART
#undef VPART_LEN
#undef VPART_SHIFT
#undef VS_CMP_AT_OFFSET
#undef VS_CMP_OF_TYPE
#undef VSKIPME
#undef VSTART
#undef VT_COMP
#undef VT_COMPNAME
#undef VT_COMPNAMES
#undef VT_COMPS
#undef VT_IDENT
#undef VT_IDENT_PTR
#undef VT_NID
#undef VT_NSUB
#undef VT_SUB
#undef VTYPE
#undef VTYPE_LEN
#undef VTYPE_SHIFT
#undef VUP
#undef VVALUE
#undef VVALUEPTR
#undef VVECSKIP
#undef VXATTR
#undef VXGHOST
#undef VXGID
#undef VXHGHOST
#undef VXMASTER
#undef VXNCOPIES
#undef VXPRIO
#undef VXPROCLIST
#undef VXPROCPRIO
#undef VXVGHOST
#undef VXVHGHOST
#undef _X_
#undef XC
#undef XCAT
#undef XCAT3
#undef XFERCOPY
#undef XFERCOPYX
#undef XFERDELETE
#undef XFERECOPY
#undef XFERECOPYX
#undef XFEREDELETE
#undef XFERMATX
#undef XFERMATX_LEN
#undef XFERMATX_SHIFT
#undef XFERVECTOR
#undef XFERVECTOR_LEN
#undef XFERVECTOR_SHIFT
#undef XFERVXCOPY
#undef XFERVXCOPYX
#undef XFERVXDELETE
#undef XI
#undef XPOS
#undef XSTR
#undef __XXL_MSIZE__
#undef _Y_
#undef YC
#undef YELLOWCLASS
#undef YES
#undef YPOS
#undef _Z_
#undef ZC
#undef ZPOS
