// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
/** \file
 * \brief Contains <tt>\#undef</tt>s for all preprocessor macros
 * defined by alberta.
 *
 * This file is created automatically by the perl script <tt>undefAllMacros.pl</tt>.
 */

#ifdef _ALBERTA_H_
  #undef _ALBERTA_H_
#endif // _ALBERTA_H_

#ifdef ALBERTA_MAGIC
  #undef ALBERTA_MAGIC
#endif // ALBERTA_MAGIC

#ifdef ALBERTA_VERSION
  #undef ALBERTA_VERSION
#endif // ALBERTA_VERSION

#ifdef DIM_LIMIT
  #undef DIM_LIMIT
#endif // DIM_LIMIT

#ifdef DIM_MAX
  #undef DIM_MAX
#endif // DIM_MAX

#ifdef N_VERTICES
  #undef N_VERTICES
#endif // N_VERTICES

#ifdef N_EDGES
  #undef N_EDGES
#endif // N_EDGES

#ifdef N_WALLS
  #undef N_WALLS
#endif // N_WALLS

#ifdef N_FACES
  #undef N_FACES
#endif // N_FACES

#ifdef N_NEIGH
  #undef N_NEIGH
#endif // N_NEIGH

#ifdef N_LAMBDA
  #undef N_LAMBDA
#endif // N_LAMBDA

#ifdef DIM_FAC
  #undef DIM_FAC
#endif // DIM_FAC

#ifdef VERTEX_OF_EDGE
  #undef VERTEX_OF_EDGE
#endif // VERTEX_OF_EDGE

#ifdef VERTEX_OF_WALL
  #undef VERTEX_OF_WALL
#endif // VERTEX_OF_WALL

#ifdef N_VERTICES_0D
  #undef N_VERTICES_0D
#endif // N_VERTICES_0D

#ifdef N_EDGES_0D
  #undef N_EDGES_0D
#endif // N_EDGES_0D

#ifdef N_FACES_0D
  #undef N_FACES_0D
#endif // N_FACES_0D

#ifdef N_NEIGH_0D
  #undef N_NEIGH_0D
#endif // N_NEIGH_0D

#ifdef N_WALLS_0D
  #undef N_WALLS_0D
#endif // N_WALLS_0D

#ifdef N_LAMBDA_0D
  #undef N_LAMBDA_0D
#endif // N_LAMBDA_0D

#ifdef DIM_FAC_0D
  #undef DIM_FAC_0D
#endif // DIM_FAC_0D

#ifdef VERTEX_OF_EDGE_0D
  #undef VERTEX_OF_EDGE_0D
#endif // VERTEX_OF_EDGE_0D

#ifdef VERTEX_OF_WALL_0D
  #undef VERTEX_OF_WALL_0D
#endif // VERTEX_OF_WALL_0D

#ifdef N_VERTICES_1D
  #undef N_VERTICES_1D
#endif // N_VERTICES_1D

#ifdef N_EDGES_1D
  #undef N_EDGES_1D
#endif // N_EDGES_1D

#ifdef N_FACES_1D
  #undef N_FACES_1D
#endif // N_FACES_1D

#ifdef N_NEIGH_1D
  #undef N_NEIGH_1D
#endif // N_NEIGH_1D

#ifdef N_WALLS_1D
  #undef N_WALLS_1D
#endif // N_WALLS_1D

#ifdef N_LAMBDA_1D
  #undef N_LAMBDA_1D
#endif // N_LAMBDA_1D

#ifdef DIM_FAC_1D
  #undef DIM_FAC_1D
#endif // DIM_FAC_1D

#ifdef VERTEX_OF_EDGE_1D
  #undef VERTEX_OF_EDGE_1D
#endif // VERTEX_OF_EDGE_1D

#ifdef VERTEX_OF_WALL_1D
  #undef VERTEX_OF_WALL_1D
#endif // VERTEX_OF_WALL_1D

#ifdef N_VERTICES_2D
  #undef N_VERTICES_2D
#endif // N_VERTICES_2D

#ifdef N_EDGES_2D
  #undef N_EDGES_2D
#endif // N_EDGES_2D

#ifdef N_FACES_2D
  #undef N_FACES_2D
#endif // N_FACES_2D

#ifdef N_NEIGH_2D
  #undef N_NEIGH_2D
#endif // N_NEIGH_2D

#ifdef N_WALLS_2D
  #undef N_WALLS_2D
#endif // N_WALLS_2D

#ifdef N_LAMBDA_2D
  #undef N_LAMBDA_2D
#endif // N_LAMBDA_2D

#ifdef DIM_FAC_2D
  #undef DIM_FAC_2D
#endif // DIM_FAC_2D

#ifdef VERTEX_OF_EDGE_2D
  #undef VERTEX_OF_EDGE_2D
#endif // VERTEX_OF_EDGE_2D

#ifdef VERTEX_OF_WALL_2D
  #undef VERTEX_OF_WALL_2D
#endif // VERTEX_OF_WALL_2D

#ifdef N_VERTICES_3D
  #undef N_VERTICES_3D
#endif // N_VERTICES_3D

#ifdef N_EDGES_3D
  #undef N_EDGES_3D
#endif // N_EDGES_3D

#ifdef N_FACES_3D
  #undef N_FACES_3D
#endif // N_FACES_3D

#ifdef N_NEIGH_3D
  #undef N_NEIGH_3D
#endif // N_NEIGH_3D

#ifdef N_WALLS_3D
  #undef N_WALLS_3D
#endif // N_WALLS_3D

#ifdef N_LAMBDA_3D
  #undef N_LAMBDA_3D
#endif // N_LAMBDA_3D

#ifdef DIM_FAC_3D
  #undef DIM_FAC_3D
#endif // DIM_FAC_3D

#ifdef VERTEX_OF_EDGE_3D
  #undef VERTEX_OF_EDGE_3D
#endif // VERTEX_OF_EDGE_3D

#ifdef VERTEX_OF_WALL_3D
  #undef VERTEX_OF_WALL_3D
#endif // VERTEX_OF_WALL_3D

#ifdef N_LAMBDA_MAX
  #undef N_LAMBDA_MAX
#endif // N_LAMBDA_MAX

#ifdef N_VERTICES_MAX
  #undef N_VERTICES_MAX
#endif // N_VERTICES_MAX

#ifdef N_EDGES_MAX
  #undef N_EDGES_MAX
#endif // N_EDGES_MAX

#ifdef N_FACES_MAX
  #undef N_FACES_MAX
#endif // N_FACES_MAX

#ifdef N_NEIGH_MAX
  #undef N_NEIGH_MAX
#endif // N_NEIGH_MAX

#ifdef N_WALLS_MAX
  #undef N_WALLS_MAX
#endif // N_WALLS_MAX

#ifdef DIM_FAC_MAX
  #undef DIM_FAC_MAX
#endif // DIM_FAC_MAX

#ifdef N_LAMBDA_LIMIT
  #undef N_LAMBDA_LIMIT
#endif // N_LAMBDA_LIMIT

#ifdef N_VERTICES_LIMIT
  #undef N_VERTICES_LIMIT
#endif // N_VERTICES_LIMIT

#ifdef N_EDGES_LIMIT
  #undef N_EDGES_LIMIT
#endif // N_EDGES_LIMIT

#ifdef N_FACES_LIMIT
  #undef N_FACES_LIMIT
#endif // N_FACES_LIMIT

#ifdef N_NEIGH_LIMIT
  #undef N_NEIGH_LIMIT
#endif // N_NEIGH_LIMIT

#ifdef N_WALLS_LIMIT
  #undef N_WALLS_LIMIT
#endif // N_WALLS_LIMIT

#ifdef DIM_FAC_LIMIT
  #undef DIM_FAC_LIMIT
#endif // DIM_FAC_LIMIT

#ifdef INDEX
  #undef INDEX
#endif // INDEX

#ifdef INDEX
  #undef INDEX
#endif // INDEX

#ifdef IS_LEAF_EL
  #undef IS_LEAF_EL
#endif // IS_LEAF_EL

#ifdef LEAF_DATA
  #undef LEAF_DATA
#endif // LEAF_DATA

#ifdef INTERIOR
  #undef INTERIOR
#endif // INTERIOR

#ifdef DIRICHLET
  #undef DIRICHLET
#endif // DIRICHLET

#ifdef NEUMANN
  #undef NEUMANN
#endif // NEUMANN

#ifdef IS_NEUMANN
  #undef IS_NEUMANN
#endif // IS_NEUMANN

#ifdef IS_DIRICHLET
  #undef IS_DIRICHLET
#endif // IS_DIRICHLET

#ifdef IS_INTERIOR
  #undef IS_INTERIOR
#endif // IS_INTERIOR

#ifdef N_BNDRY_TYPES
  #undef N_BNDRY_TYPES
#endif // N_BNDRY_TYPES

#ifdef LIST_ENTRY
  #undef LIST_ENTRY
#endif // LIST_ENTRY

#ifdef INIT_EL_TAG_CTX_INIT
  #undef INIT_EL_TAG_CTX_INIT
#endif // INIT_EL_TAG_CTX_INIT

#ifdef INIT_EL_TAG_CTX_UNIQ
  #undef INIT_EL_TAG_CTX_UNIQ
#endif // INIT_EL_TAG_CTX_UNIQ

#ifdef INIT_EL_TAG_CTX_NULL
  #undef INIT_EL_TAG_CTX_NULL
#endif // INIT_EL_TAG_CTX_NULL

#ifdef INIT_EL_TAG_CTX_DFLT
  #undef INIT_EL_TAG_CTX_DFLT
#endif // INIT_EL_TAG_CTX_DFLT

#ifdef INIT_EL_TAG_CTX_TAG
  #undef INIT_EL_TAG_CTX_TAG
#endif // INIT_EL_TAG_CTX_TAG

#ifdef INIT_ELEMENT_METHOD
  #undef INIT_ELEMENT_METHOD
#endif // INIT_ELEMENT_METHOD

#ifdef INIT_ELEMENT_FLAGS
  #undef INIT_ELEMENT_FLAGS
#endif // INIT_ELEMENT_FLAGS

#ifdef INIT_ELEMENT_DEFUN
  #undef INIT_ELEMENT_DEFUN
#endif // INIT_ELEMENT_DEFUN

#ifdef INIT_OBJECT
  #undef INIT_OBJECT
#endif // INIT_OBJECT

#ifdef INIT_ELEMENT_DECL
  #undef INIT_ELEMENT_DECL
#endif // INIT_ELEMENT_DECL

#ifdef INIT_ELEMENT_INITIALIZER
  #undef INIT_ELEMENT_INITIALIZER
#endif // INIT_ELEMENT_INITIALIZER

#ifdef INIT_ELEMENT
  #undef INIT_ELEMENT
#endif // INIT_ELEMENT

#ifdef INIT_ELEMENT_NEEDED
  #undef INIT_ELEMENT_NEEDED
#endif // INIT_ELEMENT_NEEDED

#ifdef INIT_ELEMENT_SETUP
  #undef INIT_ELEMENT_SETUP
#endif // INIT_ELEMENT_SETUP

#ifdef INIT_OBJECT_SINGLE
  #undef INIT_OBJECT_SINGLE
#endif // INIT_OBJECT_SINGLE

#ifdef INIT_ELEMENT_SINGLE
  #undef INIT_ELEMENT_SINGLE
#endif // INIT_ELEMENT_SINGLE

#ifdef vertex_of_wall_2d
  #undef vertex_of_wall_2d
#endif // vertex_of_wall_2d

#ifdef PARAM_STRATEGY_MASK
  #undef PARAM_STRATEGY_MASK
#endif // PARAM_STRATEGY_MASK

#ifdef PARAM_PERIODIC_COORDS
  #undef PARAM_PERIODIC_COORDS
#endif // PARAM_PERIODIC_COORDS

#ifdef FILL_EL_DET
  #undef FILL_EL_DET
#endif // FILL_EL_DET

#ifdef FILL_EL_LAMBDA
  #undef FILL_EL_LAMBDA
#endif // FILL_EL_LAMBDA

#ifdef FILL_EL_WALL_SHIFT
  #undef FILL_EL_WALL_SHIFT
#endif // FILL_EL_WALL_SHIFT

#ifdef FILL_EL_WALL_MASK
  #undef FILL_EL_WALL_MASK
#endif // FILL_EL_WALL_MASK

#ifdef FILL_EL_WALL_DET
  #undef FILL_EL_WALL_DET
#endif // FILL_EL_WALL_DET

#ifdef FILL_EL_WALL_NORMAL
  #undef FILL_EL_WALL_NORMAL
#endif // FILL_EL_WALL_NORMAL

#ifdef FILL_EL_WALL_ORIENTATION
  #undef FILL_EL_WALL_ORIENTATION
#endif // FILL_EL_WALL_ORIENTATION

#ifdef FILL_EL_WALL_REL_ORIENTATION
  #undef FILL_EL_WALL_REL_ORIENTATION
#endif // FILL_EL_WALL_REL_ORIENTATION

#ifdef FILL_EL_WALL_DETS
  #undef FILL_EL_WALL_DETS
#endif // FILL_EL_WALL_DETS

#ifdef FILL_EL_WALL_NORMALS
  #undef FILL_EL_WALL_NORMALS
#endif // FILL_EL_WALL_NORMALS

#ifdef FILL_EL_WALL_ORIENTATIONS
  #undef FILL_EL_WALL_ORIENTATIONS
#endif // FILL_EL_WALL_ORIENTATIONS

#ifdef FILL_EL_WALL_REL_ORIENTATIONS
  #undef FILL_EL_WALL_REL_ORIENTATIONS
#endif // FILL_EL_WALL_REL_ORIENTATIONS

#ifdef BNDRY_FLAGS_INIT
  #undef BNDRY_FLAGS_INIT
#endif // BNDRY_FLAGS_INIT

#ifdef BNDRY_FLAGS_ALL
  #undef BNDRY_FLAGS_ALL
#endif // BNDRY_FLAGS_ALL

#ifdef BNDRY_FLAGS_CPY
  #undef BNDRY_FLAGS_CPY
#endif // BNDRY_FLAGS_CPY

#ifdef BNDRY_FLAGS_AND
  #undef BNDRY_FLAGS_AND
#endif // BNDRY_FLAGS_AND

#ifdef BNDRY_FLAGS_OR
  #undef BNDRY_FLAGS_OR
#endif // BNDRY_FLAGS_OR

#ifdef BNDRY_FLAGS_XOR
  #undef BNDRY_FLAGS_XOR
#endif // BNDRY_FLAGS_XOR

#ifdef BNDRY_FLAGS_IS_INTERIOR
  #undef BNDRY_FLAGS_IS_INTERIOR
#endif // BNDRY_FLAGS_IS_INTERIOR

#ifdef BNDRY_FLAGS_IS_AT_BNDRY
  #undef BNDRY_FLAGS_IS_AT_BNDRY
#endif // BNDRY_FLAGS_IS_AT_BNDRY

#ifdef BNDRY_FLAGS_SET
  #undef BNDRY_FLAGS_SET
#endif // BNDRY_FLAGS_SET

#ifdef BNDRY_FLAGS_IS_PARTOF
  #undef BNDRY_FLAGS_IS_PARTOF
#endif // BNDRY_FLAGS_IS_PARTOF

#ifdef BNDRY_FLAGS_FFBB
  #undef BNDRY_FLAGS_FFBB
#endif // BNDRY_FLAGS_FFBB

#ifdef RCLE_NONE
  #undef RCLE_NONE
#endif // RCLE_NONE

#ifdef RCLE_COARSE_EDGE_COMPAT
  #undef RCLE_COARSE_EDGE_COMPAT
#endif // RCLE_COARSE_EDGE_COMPAT

#ifdef FILL_NOTHING
  #undef FILL_NOTHING
#endif // FILL_NOTHING

#ifdef FILL_COORDS
  #undef FILL_COORDS
#endif // FILL_COORDS

#ifdef FILL_BOUND
  #undef FILL_BOUND
#endif // FILL_BOUND

#ifdef FILL_NEIGH
  #undef FILL_NEIGH
#endif // FILL_NEIGH

#ifdef FILL_OPP_COORDS
  #undef FILL_OPP_COORDS
#endif // FILL_OPP_COORDS

#ifdef FILL_ORIENTATION
  #undef FILL_ORIENTATION
#endif // FILL_ORIENTATION

#ifdef FILL_PROJECTION
  #undef FILL_PROJECTION
#endif // FILL_PROJECTION

#ifdef FILL_MACRO_WALLS
  #undef FILL_MACRO_WALLS
#endif // FILL_MACRO_WALLS

#ifdef FILL_WALL_MAP
  #undef FILL_WALL_MAP
#endif // FILL_WALL_MAP

#ifdef FILL_NON_PERIODIC
  #undef FILL_NON_PERIODIC
#endif // FILL_NON_PERIODIC

#ifdef FILL_MASTER_INFO
  #undef FILL_MASTER_INFO
#endif // FILL_MASTER_INFO

#ifdef FILL_ANY
  #undef FILL_ANY
#endif // FILL_ANY

#ifdef CALL_EVERY_EL_PREORDER
  #undef CALL_EVERY_EL_PREORDER
#endif // CALL_EVERY_EL_PREORDER

#ifdef CALL_EVERY_EL_INORDER
  #undef CALL_EVERY_EL_INORDER
#endif // CALL_EVERY_EL_INORDER

#ifdef CALL_EVERY_EL_POSTORDER
  #undef CALL_EVERY_EL_POSTORDER
#endif // CALL_EVERY_EL_POSTORDER

#ifdef CALL_LEAF_EL
  #undef CALL_LEAF_EL
#endif // CALL_LEAF_EL

#ifdef CALL_LEAF_EL_LEVEL
  #undef CALL_LEAF_EL_LEVEL
#endif // CALL_LEAF_EL_LEVEL

#ifdef CALL_EL_LEVEL
  #undef CALL_EL_LEVEL
#endif // CALL_EL_LEVEL

#ifdef CALL_MG_LEVEL
  #undef CALL_MG_LEVEL
#endif // CALL_MG_LEVEL

#ifdef TEST_FLAG
  #undef TEST_FLAG
#endif // TEST_FLAG

#ifdef ADM_FLAGS_MASK
  #undef ADM_FLAGS_MASK
#endif // ADM_FLAGS_MASK

#ifdef UCHAR_name
  #undef UCHAR_name
#endif // UCHAR_name

#ifdef uchar_VECNAME
  #undef uchar_VECNAME
#endif // uchar_VECNAME

#ifdef SCHAR_name
  #undef SCHAR_name
#endif // SCHAR_name

#ifdef schar_VECNAME
  #undef schar_VECNAME
#endif // schar_VECNAME

#ifdef INT_name
  #undef INT_name
#endif // INT_name

#ifdef int_VECNAME
  #undef int_VECNAME
#endif // int_VECNAME

#ifdef DOF_name
  #undef DOF_name
#endif // DOF_name

#ifdef dof_VECNAME
  #undef dof_VECNAME
#endif // dof_VECNAME

#ifdef PTR_name
  #undef PTR_name
#endif // PTR_name

#ifdef ptr_VECNAME
  #undef ptr_VECNAME
#endif // ptr_VECNAME

#ifdef REAL_name
  #undef REAL_name
#endif // REAL_name

#ifdef real_VECNAME
  #undef real_VECNAME
#endif // real_VECNAME

#ifdef REAL_D_name
  #undef REAL_D_name
#endif // REAL_D_name

#ifdef real_d_VECNAME
  #undef real_d_VECNAME
#endif // real_d_VECNAME

#ifdef BNDRY_name
  #undef BNDRY_name
#endif // BNDRY_name

#ifdef bndry_VECNAME
  #undef bndry_VECNAME
#endif // bndry_VECNAME

#ifdef DECL_DOF_EL_VEC
  #undef DECL_DOF_EL_VEC
#endif // DECL_DOF_EL_VEC

#ifdef PHI
  #undef PHI
#endif // PHI

#ifdef GRD_PHI
  #undef GRD_PHI
#endif // GRD_PHI

#ifdef D2_PHI
  #undef D2_PHI
#endif // D2_PHI

#ifdef D3_PHI
  #undef D3_PHI
#endif // D3_PHI

#ifdef D4_PHI
  #undef D4_PHI
#endif // D4_PHI

#ifdef PHI_D
  #undef PHI_D
#endif // PHI_D

#ifdef GRD_PHI_D
  #undef GRD_PHI_D
#endif // GRD_PHI_D

#ifdef D2_PHI_D
  #undef D2_PHI_D
#endif // D2_PHI_D

#ifdef GET_DOF_INDICES
  #undef GET_DOF_INDICES
#endif // GET_DOF_INDICES

#ifdef INTERPOL
  #undef INTERPOL
#endif // INTERPOL

#ifdef INTERPOL_D
  #undef INTERPOL_D
#endif // INTERPOL_D

#ifdef INTERPOL_DOW
  #undef INTERPOL_DOW
#endif // INTERPOL_DOW

#ifdef GET_BOUND
  #undef GET_BOUND
#endif // GET_BOUND

#ifdef LAGRANGE_NODES
  #undef LAGRANGE_NODES
#endif // LAGRANGE_NODES

#ifdef FILL_EL_QUAD_WORLD
  #undef FILL_EL_QUAD_WORLD
#endif // FILL_EL_QUAD_WORLD

#ifdef FILL_EL_QUAD_DET
  #undef FILL_EL_QUAD_DET
#endif // FILL_EL_QUAD_DET

#ifdef FILL_EL_QUAD_LAMBDA
  #undef FILL_EL_QUAD_LAMBDA
#endif // FILL_EL_QUAD_LAMBDA

#ifdef FILL_EL_QUAD_DLAMBDA
  #undef FILL_EL_QUAD_DLAMBDA
#endif // FILL_EL_QUAD_DLAMBDA

#ifdef FILL_EL_QUAD_GRD_WORLD
  #undef FILL_EL_QUAD_GRD_WORLD
#endif // FILL_EL_QUAD_GRD_WORLD

#ifdef FILL_EL_QUAD_D2_WORLD
  #undef FILL_EL_QUAD_D2_WORLD
#endif // FILL_EL_QUAD_D2_WORLD

#ifdef FILL_EL_QUAD_D3_WORLD
  #undef FILL_EL_QUAD_D3_WORLD
#endif // FILL_EL_QUAD_D3_WORLD

#ifdef FILL_EL_QUAD_WALL_DET
  #undef FILL_EL_QUAD_WALL_DET
#endif // FILL_EL_QUAD_WALL_DET

#ifdef FILL_EL_QUAD_WALL_NORMAL
  #undef FILL_EL_QUAD_WALL_NORMAL
#endif // FILL_EL_QUAD_WALL_NORMAL

#ifdef FILL_EL_QUAD_GRD_NORMAL
  #undef FILL_EL_QUAD_GRD_NORMAL
#endif // FILL_EL_QUAD_GRD_NORMAL

#ifdef FILL_EL_QUAD_D2_NORMAL
  #undef FILL_EL_QUAD_D2_NORMAL
#endif // FILL_EL_QUAD_D2_NORMAL

#ifdef INIT_PHI
  #undef INIT_PHI
#endif // INIT_PHI

#ifdef INIT_GRD_PHI
  #undef INIT_GRD_PHI
#endif // INIT_GRD_PHI

#ifdef INIT_D2_PHI
  #undef INIT_D2_PHI
#endif // INIT_D2_PHI

#ifdef INIT_D3_PHI
  #undef INIT_D3_PHI
#endif // INIT_D3_PHI

#ifdef INIT_D4_PHI
  #undef INIT_D4_PHI
#endif // INIT_D4_PHI

#ifdef INIT_TANGENTIAL
  #undef INIT_TANGENTIAL
#endif // INIT_TANGENTIAL

#ifdef MESH_REFINED
  #undef MESH_REFINED
#endif // MESH_REFINED

#ifdef MESH_COARSENED
  #undef MESH_COARSENED
#endif // MESH_COARSENED

#ifdef GRAPH_MESH_BOUNDARY
  #undef GRAPH_MESH_BOUNDARY
#endif // GRAPH_MESH_BOUNDARY

#ifdef GRAPH_MESH_ELEMENT_MARK
  #undef GRAPH_MESH_ELEMENT_MARK
#endif // GRAPH_MESH_ELEMENT_MARK

#ifdef GRAPH_MESH_VERTEX_DOF
  #undef GRAPH_MESH_VERTEX_DOF
#endif // GRAPH_MESH_VERTEX_DOF

#ifdef GRAPH_MESH_ELEMENT_INDEX
  #undef GRAPH_MESH_ELEMENT_INDEX
#endif // GRAPH_MESH_ELEMENT_INDEX

#ifdef GET_MESH
  #undef GET_MESH
#endif // GET_MESH

#ifdef GET_DOF_VEC
  #undef GET_DOF_VEC
#endif // GET_DOF_VEC

#ifdef TRACE_DOF_VEC_PROTO
  #undef TRACE_DOF_VEC_PROTO
#endif // TRACE_DOF_VEC_PROTO

#ifdef INIT_UH
  #undef INIT_UH
#endif // INIT_UH

#ifdef INIT_GRD_UH
  #undef INIT_GRD_UH
#endif // INIT_GRD_UH

#ifdef ellipt_est_d_init
  #undef ellipt_est_d_init
#endif // ellipt_est_d_init

#ifdef head_est_d_init
  #undef head_est_d_init
#endif // head_est_d_init

#ifdef element_est_d
  #undef element_est_d
#endif // element_est_d

#ifdef element_est_d_finish
  #undef element_est_d_finish
#endif // element_est_d_finish

#ifdef element_est_uh_d
  #undef element_est_uh_d
#endif // element_est_uh_d

#ifdef element_est_grd_uh_d
  #undef element_est_grd_uh_d
#endif // element_est_grd_uh_d

#ifdef ellipt_est_d_finish
  #undef ellipt_est_d_finish
#endif // ellipt_est_d_finish

#ifdef heat_est_d_finish
  #undef heat_est_d_finish
#endif // heat_est_d_finish

#ifdef TRAVERSE_FIRST
  #undef TRAVERSE_FIRST
#endif // TRAVERSE_FIRST

#ifdef TRAVERSE_NEXT
  #undef TRAVERSE_NEXT
#endif // TRAVERSE_NEXT

#ifdef TRAVERSE_START
  #undef TRAVERSE_START
#endif // TRAVERSE_START

#ifdef TRAVERSE_STOP
  #undef TRAVERSE_STOP
#endif // TRAVERSE_STOP

#ifdef TRAVERSE_NEIGHBOUR
  #undef TRAVERSE_NEIGHBOUR
#endif // TRAVERSE_NEIGHBOUR

#ifdef IFF_TAG_ALBERTA
  #undef IFF_TAG_ALBERTA
#endif // IFF_TAG_ALBERTA

#ifdef IFF_TAG_MESH
  #undef IFF_TAG_MESH
#endif // IFF_TAG_MESH

#ifdef IFF_TAG_REAL_VEC
  #undef IFF_TAG_REAL_VEC
#endif // IFF_TAG_REAL_VEC

#ifdef IFF_TAG_REAL_D_VEC
  #undef IFF_TAG_REAL_D_VEC
#endif // IFF_TAG_REAL_D_VEC

#ifdef IFF_TAG_INT_VEC
  #undef IFF_TAG_INT_VEC
#endif // IFF_TAG_INT_VEC

#ifdef IFF_TAG_UCHAR_VEC
  #undef IFF_TAG_UCHAR_VEC
#endif // IFF_TAG_UCHAR_VEC

#ifdef IFF_TAG_SCHAR_VEC
  #undef IFF_TAG_SCHAR_VEC
#endif // IFF_TAG_SCHAR_VEC
