// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_DGFPARSER_HH
#define DUNE_DGFPARSER_HH
// include dgf parser
#include "dgfparser/dgfparser.hh"
/* include the implementations */
#include <dune/grid/albertagrid/dgfparser.hh>
#include "dgfparser/dgfug.hh"
#include "dgfparser/dgfoned.hh"
#include "dgfparser/dgfyasp.hh"
#include "dgfparser/dgfgeogrid.hh"
#include "dgfparser/dgfidentitygrid.hh"
#endif
