// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_GEOGRID_DECLARATION_HH
#define DUNE_GEOGRID_DECLARATION_HH

namespace Dune
{

  template< class HostGrid, class CoordFunction, class Allocator >
  class GeometryGrid;

} // namespace Dune

#endif // #ifndef DUNE_GEOGRID_DECLARATION_HH
