// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
/** \file
 * \brief Contains <tt>\#undef</tt>s for all preprocessor macros
 * defined by alberta.
 *
 * This file is created automatically by the perl script <tt>undefAllMacros.pl</tt>.
 */

#ifdef _ALBERTA_H_
  #undef _ALBERTA_H_
#endif // _ALBERTA_H_

#ifdef ALBERTA_VERSION
  #undef ALBERTA_VERSION
#endif // ALBERTA_VERSION

#ifdef ALBERTA_DEBUG
  #undef ALBERTA_DEBUG
#endif // ALBERTA_DEBUG

#ifdef DIM_MAX
  #undef DIM_MAX
#endif // DIM_MAX

#ifdef N_VERTICES
  #undef N_VERTICES
#endif // N_VERTICES

#ifdef N_EDGES
  #undef N_EDGES
#endif // N_EDGES

#ifdef N_WALLS
  #undef N_WALLS
#endif // N_WALLS

#ifdef N_FACES
  #undef N_FACES
#endif // N_FACES

#ifdef N_NEIGH
  #undef N_NEIGH
#endif // N_NEIGH

#ifdef DIM_FAC
  #undef DIM_FAC
#endif // DIM_FAC

#ifdef N_VERTICES_0D
  #undef N_VERTICES_0D
#endif // N_VERTICES_0D

#ifdef N_EDGES_0D
  #undef N_EDGES_0D
#endif // N_EDGES_0D

#ifdef N_FACES_0D
  #undef N_FACES_0D
#endif // N_FACES_0D

#ifdef N_NEIGH_0D
  #undef N_NEIGH_0D
#endif // N_NEIGH_0D

#ifdef N_WALLS_0D
  #undef N_WALLS_0D
#endif // N_WALLS_0D

#ifdef N_VERTICES_1D
  #undef N_VERTICES_1D
#endif // N_VERTICES_1D

#ifdef N_EDGES_1D
  #undef N_EDGES_1D
#endif // N_EDGES_1D

#ifdef N_FACES_1D
  #undef N_FACES_1D
#endif // N_FACES_1D

#ifdef N_NEIGH_1D
  #undef N_NEIGH_1D
#endif // N_NEIGH_1D

#ifdef N_WALLS_1D
  #undef N_WALLS_1D
#endif // N_WALLS_1D

#ifdef N_VERTICES_2D
  #undef N_VERTICES_2D
#endif // N_VERTICES_2D

#ifdef N_EDGES_2D
  #undef N_EDGES_2D
#endif // N_EDGES_2D

#ifdef N_FACES_2D
  #undef N_FACES_2D
#endif // N_FACES_2D

#ifdef N_NEIGH_2D
  #undef N_NEIGH_2D
#endif // N_NEIGH_2D

#ifdef N_WALLS_2D
  #undef N_WALLS_2D
#endif // N_WALLS_2D

#ifdef N_VERTICES_3D
  #undef N_VERTICES_3D
#endif // N_VERTICES_3D

#ifdef N_EDGES_3D
  #undef N_EDGES_3D
#endif // N_EDGES_3D

#ifdef N_FACES_3D
  #undef N_FACES_3D
#endif // N_FACES_3D

#ifdef N_NEIGH_3D
  #undef N_NEIGH_3D
#endif // N_NEIGH_3D

#ifdef N_WALLS_3D
  #undef N_WALLS_3D
#endif // N_WALLS_3D

#ifdef N_VERTICES_MAX
  #undef N_VERTICES_MAX
#endif // N_VERTICES_MAX

#ifdef N_EDGES_MAX
  #undef N_EDGES_MAX
#endif // N_EDGES_MAX

#ifdef N_FACES_MAX
  #undef N_FACES_MAX
#endif // N_FACES_MAX

#ifdef N_NEIGH_MAX
  #undef N_NEIGH_MAX
#endif // N_NEIGH_MAX

#ifdef N_WALLS_MAX
  #undef N_WALLS_MAX
#endif // N_WALLS_MAX

#ifdef N_LAMBDA
  #undef N_LAMBDA
#endif // N_LAMBDA

#ifdef LAMBDA_MAX
  #undef LAMBDA_MAX
#endif // LAMBDA_MAX

#ifdef INDEX
  #undef INDEX
#endif // INDEX

#ifdef INDEX
  #undef INDEX
#endif // INDEX

#ifdef IS_LEAF_EL
  #undef IS_LEAF_EL
#endif // IS_LEAF_EL

#ifdef LEAF_DATA
  #undef LEAF_DATA
#endif // LEAF_DATA

#ifdef INTERIOR
  #undef INTERIOR
#endif // INTERIOR

#ifdef DIRICHLET
  #undef DIRICHLET
#endif // DIRICHLET

#ifdef NEUMANN
  #undef NEUMANN
#endif // NEUMANN

#ifdef IS_NEUMANN
  #undef IS_NEUMANN
#endif // IS_NEUMANN

#ifdef IS_DIRICHLET
  #undef IS_DIRICHLET
#endif // IS_DIRICHLET

#ifdef IS_INTERIOR
  #undef IS_INTERIOR
#endif // IS_INTERIOR

#ifdef N_NODE_TYPES
  #undef N_NODE_TYPES
#endif // N_NODE_TYPES

#ifdef VERTEX
  #undef VERTEX
#endif // VERTEX

#ifdef CENTER
  #undef CENTER
#endif // CENTER

#ifdef EDGE
  #undef EDGE
#endif // EDGE

#ifdef FACE
  #undef FACE
#endif // FACE

#ifdef LIST_ENTRY
  #undef LIST_ENTRY
#endif // LIST_ENTRY

#ifdef WALL_BOUND
  #undef WALL_BOUND
#endif // WALL_BOUND

#ifdef FILL_NOTHING
  #undef FILL_NOTHING
#endif // FILL_NOTHING

#ifdef FILL_COORDS
  #undef FILL_COORDS
#endif // FILL_COORDS

#ifdef FILL_BOUND
  #undef FILL_BOUND
#endif // FILL_BOUND

#ifdef FILL_NEIGH
  #undef FILL_NEIGH
#endif // FILL_NEIGH

#ifdef FILL_OPP_COORDS
  #undef FILL_OPP_COORDS
#endif // FILL_OPP_COORDS

#ifdef FILL_ORIENTATION
  #undef FILL_ORIENTATION
#endif // FILL_ORIENTATION

#ifdef FILL_EL_TYPE
  #undef FILL_EL_TYPE
#endif // FILL_EL_TYPE

#ifdef FILL_PROJECTION
  #undef FILL_PROJECTION
#endif // FILL_PROJECTION

#ifdef FILL_ANY
  #undef FILL_ANY
#endif // FILL_ANY

#ifdef HOOK_QUEUE_OFFSET
  #undef HOOK_QUEUE_OFFSET
#endif // HOOK_QUEUE_OFFSET

#ifdef RUN_HOOKS_0
  #undef RUN_HOOKS_0
#endif // RUN_HOOKS_0

#ifdef RUN_BAS_FCTS_HOOKS
  #undef RUN_BAS_FCTS_HOOKS
#endif // RUN_BAS_FCTS_HOOKS

#ifdef RUN_HOOKS_1
  #undef RUN_HOOKS_1
#endif // RUN_HOOKS_1

#ifdef RUN_HOOKS_2
  #undef RUN_HOOKS_2
#endif // RUN_HOOKS_2

#ifdef RUN_HOOKS_3
  #undef RUN_HOOKS_3
#endif // RUN_HOOKS_3

#ifdef RUN_HOOKS_4
  #undef RUN_HOOKS_4
#endif // RUN_HOOKS_4

#ifdef RUN_HOOKS_5
  #undef RUN_HOOKS_5
#endif // RUN_HOOKS_5

#ifdef RUN_HOOKS_6
  #undef RUN_HOOKS_6
#endif // RUN_HOOKS_6

#ifdef RUN_HOOKS_7
  #undef RUN_HOOKS_7
#endif // RUN_HOOKS_7

#ifdef CALL_EVERY_EL_PREORDER
  #undef CALL_EVERY_EL_PREORDER
#endif // CALL_EVERY_EL_PREORDER

#ifdef CALL_EVERY_EL_INORDER
  #undef CALL_EVERY_EL_INORDER
#endif // CALL_EVERY_EL_INORDER

#ifdef CALL_EVERY_EL_POSTORDER
  #undef CALL_EVERY_EL_POSTORDER
#endif // CALL_EVERY_EL_POSTORDER

#ifdef CALL_LEAF_EL
  #undef CALL_LEAF_EL
#endif // CALL_LEAF_EL

#ifdef CALL_LEAF_EL_LEVEL
  #undef CALL_LEAF_EL_LEVEL
#endif // CALL_LEAF_EL_LEVEL

#ifdef CALL_EL_LEVEL
  #undef CALL_EL_LEVEL
#endif // CALL_EL_LEVEL

#ifdef CALL_MG_LEVEL
  #undef CALL_MG_LEVEL
#endif // CALL_MG_LEVEL

#ifdef TEST_FLAG
  #undef TEST_FLAG
#endif // TEST_FLAG

#ifdef DOF_ADMIN_DEF
  #undef DOF_ADMIN_DEF
#endif // DOF_ADMIN_DEF

#ifdef DOF_FREE_SIZE
  #undef DOF_FREE_SIZE
#endif // DOF_FREE_SIZE

#ifdef DOF_UNIT_ALL_FREE
  #undef DOF_UNIT_ALL_FREE
#endif // DOF_UNIT_ALL_FREE

#ifdef FOR_ALL_DOFS
  #undef FOR_ALL_DOFS
#endif // FOR_ALL_DOFS

#ifdef FOR_ALL_FREE_DOFS
  #undef FOR_ALL_FREE_DOFS
#endif // FOR_ALL_FREE_DOFS

#ifdef DOWBM_FULL
  #undef DOWBM_FULL
#endif // DOWBM_FULL

#ifdef DOWBM_SYMM
  #undef DOWBM_SYMM
#endif // DOWBM_SYMM

#ifdef DOWBM_DIAG
  #undef DOWBM_DIAG
#endif // DOWBM_DIAG

#ifdef ROW_LENGTH
  #undef ROW_LENGTH
#endif // ROW_LENGTH

#ifdef ENTRY_USED
  #undef ENTRY_USED
#endif // ENTRY_USED

#ifdef ENTRY_NOT_USED
  #undef ENTRY_NOT_USED
#endif // ENTRY_NOT_USED

#ifdef UNUSED_ENTRY
  #undef UNUSED_ENTRY
#endif // UNUSED_ENTRY

#ifdef NO_MORE_ENTRIES
  #undef NO_MORE_ENTRIES
#endif // NO_MORE_ENTRIES

#ifdef LAGRANGE_NODES
  #undef LAGRANGE_NODES
#endif // LAGRANGE_NODES

#ifdef MAX_N_QUAD_POINTS
  #undef MAX_N_QUAD_POINTS
#endif // MAX_N_QUAD_POINTS

#ifdef INIT_PHI
  #undef INIT_PHI
#endif // INIT_PHI

#ifdef INIT_GRD_PHI
  #undef INIT_GRD_PHI
#endif // INIT_GRD_PHI

#ifdef INIT_D2_PHI
  #undef INIT_D2_PHI
#endif // INIT_D2_PHI

#ifdef MESH_REFINED
  #undef MESH_REFINED
#endif // MESH_REFINED

#ifdef MESH_COARSENED
  #undef MESH_COARSENED
#endif // MESH_COARSENED

#ifdef H1_NORM
  #undef H1_NORM
#endif // H1_NORM

#ifdef L2_NORM
  #undef L2_NORM
#endif // L2_NORM

#ifdef GRAPH_MESH_BOUNDARY
  #undef GRAPH_MESH_BOUNDARY
#endif // GRAPH_MESH_BOUNDARY

#ifdef GRAPH_MESH_ELEMENT_MARK
  #undef GRAPH_MESH_ELEMENT_MARK
#endif // GRAPH_MESH_ELEMENT_MARK

#ifdef GRAPH_MESH_VERTEX_DOF
  #undef GRAPH_MESH_VERTEX_DOF
#endif // GRAPH_MESH_VERTEX_DOF

#ifdef GRAPH_MESH_ELEMENT_INDEX
  #undef GRAPH_MESH_ELEMENT_INDEX
#endif // GRAPH_MESH_ELEMENT_INDEX

#ifdef GET_MESH
  #undef GET_MESH
#endif // GET_MESH

#ifdef GET_DOF_VEC
  #undef GET_DOF_VEC
#endif // GET_DOF_VEC

#ifdef CARD_SN_MAX
  #undef CARD_SN_MAX
#endif // CARD_SN_MAX

#ifdef INIT_UH
  #undef INIT_UH
#endif // INIT_UH

#ifdef INIT_GRD_UH
  #undef INIT_GRD_UH
#endif // INIT_GRD_UH

#ifdef TRAVERSE_START
  #undef TRAVERSE_START
#endif // TRAVERSE_START

#ifdef TRAVERSE_STOP
  #undef TRAVERSE_STOP
#endif // TRAVERSE_STOP
